#!/usr/bin/env python3
import unittest
from unittest import mock
from unittest.mock import patch, MagicMock
import io
import urllib
import json
import os
import time

import pirate.torrent
import pirate.data
from pirate.print import Printer
from tests import util


class TestTorrent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # to make test deterministic
        os.environ['TZ'] = 'Etc/UTC'
        time.tzset()

    def test_no_hits(self):
        expected = []
        with util.open_data('no_hits.json') as res:
            actual = pirate.torrent.parse_page(res)
        self.assertEqual(actual, expected)

    def test_blocked_mirror(self):
        with util.open_data('blocked.html') as res:
            with self.assertRaises(IOError):
                pirate.torrent.parse_page(res)

    def test_search_results(self):
        with util.open_data('result.json') as file:
            expected = json.load(file)
        with util.open_data('debian_iso.json') as res:
            actual = pirate.torrent.parse_page(res)
        json.dump(actual, open('result.json', 'w'))
        self.assertEqual(actual, expected)

    def test_parse_category(self):
        category = pirate.torrent.parse_category(MagicMock(Printer), 'Audio')
        self.assertEqual(100, category)
        category = pirate.torrent.parse_category(MagicMock(Printer), 'Video')
        self.assertEqual(200, category)
        category = pirate.torrent.parse_category(MagicMock(Printer), '100')
        self.assertEqual(100, category)
        category = pirate.torrent.parse_category(MagicMock(Printer), 'asdf')
        self.assertEqual(0, category)
        category = pirate.torrent.parse_category(MagicMock(Printer), '9001')
        self.assertEqual(0, category)

    def test_parse_sort(self):
        sort = pirate.torrent.parse_sort(MagicMock(Printer), 'SeedersDsc')
        self.assertEqual(['seeders', True], sort)
        sort = pirate.torrent.parse_sort(MagicMock(Printer), 'CategoryAsc')
        self.assertEqual(['category', False], sort)
        sort = pirate.torrent.parse_sort(MagicMock(Printer), 'DateAsc')
        self.assertEqual(['raw_uploaded', False], sort)
        sort = pirate.torrent.parse_sort(MagicMock(Printer), '7')
        self.assertEqual(['seeders', True], sort)
        sort = pirate.torrent.parse_sort(MagicMock(Printer), 'asdf')
        self.assertEqual(['seeders', True], sort)
        sort = pirate.torrent.parse_sort(MagicMock(Printer), '7000')
        self.assertEqual(['seeders', True], sort)

    def test_request_path(self):
        # the args are (mode, category, terms)
        succeed = [
            (('recent', 1,   0, []), '/precompiled/data_top100_recent_1.json'),
            (('recent', 2, 100, []), '/precompiled/data_top100_recent_2.json'),
            (('top',    1,   0, []), '/precompiled/data_top100_all.json'),
            (('top',    1, 100, []), '/precompiled/data_top100_100.json'),
            (('search', 1, 100, ['abc']), '/q.php?q=abc&cat=100'),
            (('search', 1, 100, ['abc', 'def']), '/q.php?q=abc%20def&cat=100'),
            (('search', 1, 100, ['\u1234']), '/q.php?q=%E1%88%B4&cat=100'),
            (('browse', 1, 100, []), '/q.php?q=category%3A100'),
        ]
        fail = [
            (('browse', 1,   0, []), Exception),
            (('asdf',   1, 100, []), Exception)
        ]
        for inp, out in succeed:
            path = pirate.torrent.build_request_path(*inp)
            self.assertEqual(out, path)
        for inp, out, in fail:
            with self.assertRaises(out):
                pirate.torrent.build_request_path(*inp)

    @patch('pirate.torrent.get_torrent')
    def test_save_torrents(self, get_torrent):
        with patch('pirate.torrent.open',
                   mock.mock_open(), create=True) as open_:
            pirate.torrent.save_torrents(
                MagicMock(Printer), [0],
                [{'name': 'cool torrent',
                  'info_hash': 3735928559,
                  'magnet': 'magnet:?xt=urn:btih:deadbeef'}], 'path', 9)
            get_torrent.assert_called_once_with(3735928559, 9)
            open_.assert_called_once_with('path/cool torrent.torrent', 'wb')

    @patch('pirate.torrent.get_torrent',
           side_effect=urllib.error.HTTPError('', '', '', '', io.StringIO()))
    def test_save_torrents_fail(self, get_torrent):
        pirate.torrent.save_torrents(
            MagicMock(Printer), [0],
            [{'name': 'cool torrent',
              'info_hash': 3735928559,
              'magnet': 'magnet:?xt=urn:btih:deadbeef'}], 'path', 9)

    def test_save_magnets(self):
        with patch('pirate.torrent.open',
                   mock.mock_open(), create=True) as open_:
            pirate.torrent.save_magnets(
                MagicMock(Printer), [0],
                [{'name': 'cool torrent',
                  'info_hash': 3735928559,
                  'magnet': 'magnet:?xt=urn:btih:deadbeef'}], 'path')
            open_.assert_called_once_with('path/cool torrent.magnet', 'w')

    @patch('urllib.request.urlopen')
    def test_get_torrent(self, urlopen):
        class MockRequest():
            add_header = mock.MagicMock()
        request_obj = MockRequest()
        with patch('urllib.request.Request', return_value=request_obj) as req:
            pirate.torrent.get_torrent(100000000000000, 9)
            req.assert_called_once_with(
                'http://itorrents.org/torrent/5AF3107A4000.torrent',
                headers=pirate.data.default_headers)
            urlopen.assert_called_once_with(
                request_obj,
                timeout=9)

    def test_remote(self):
        class MockRequest():
            add_header = mock.MagicMock()
        req_obj = MockRequest()

        class MockInfo():
            get_content_type = mock.MagicMock(return_value='application/json')
            get = mock.MagicMock()

        class MockResponse():
            read = mock.MagicMock(return_value=b'[]')
            info = mock.MagicMock(return_value=MockInfo())
        res_obj = MockResponse()

        sort = pirate.torrent.parse_sort(MagicMock(Printer), 10)

        with patch('urllib.request.Request', return_value=req_obj) as req:
            with patch('urllib.request.urlopen', return_value=res_obj) as res:
                results = pirate.torrent.remote(
                    MagicMock(Printer), 1, 100, sort, 'top',
                    [], 'http://example.com', 9)
                req.assert_called_once_with(
                    'http://example.com/precompiled/data_top100_100.json',
                    headers=pirate.data.default_headers)
                res.assert_called_once_with(req_obj, timeout=9)
                self.assertEqual(results, [])


if __name__ == '__main__':
    unittest.main()



import builtins
import re
import gzip
import urllib.request as request
import shutil
import json
import sys

import pirate.data
import pirate.torrent

import colorama
import veryprettytable as pretty

from io import BytesIO


class Printer:
    def __init__(self, enable_color):
        self.enable_color = enable_color

    def print(self, *args, **kwargs):
        if kwargs.get('color', False) and self.enable_color:
            colorama.init()
            color_dict = {
                'default': '',
                'header':  colorama.Back.BLACK + colorama.Fore.WHITE,
                'alt':     colorama.Fore.YELLOW,
                'zebra_0': '',
                'zebra_1': colorama.Fore.BLUE,
                'WARN':    colorama.Fore.MAGENTA,
                'ERROR':   colorama.Fore.RED}

            c = color_dict[kwargs.pop('color')]
            args = (c + args[0],) + args[1:] + (colorama.Style.RESET_ALL,)
            kwargs.pop('color', None)
            return builtins.print(*args, file=sys.stderr, **kwargs)
        else:
            kwargs.pop('color', None)
            return builtins.print(*args, file=sys.stderr, **kwargs)

    # TODO: extract the name from the search results
    #       instead of from the magnet link when possible
    def search_results(self, results, local=None):
        columns = shutil.get_terminal_size((80, 20)).columns
        even = True

        if local:
            table = pretty.VeryPrettyTable(['LINK', 'DATE', 'SIZE', 'NAME'])

            table.align['SIZE'] = 'r'
            table.align['NAME'] = 'l'
        else:
            table = pretty.VeryPrettyTable(['LINK', 'SEED', 'LEECH',
                                            'RATIO', 'SIZE',
                                            'UPLOAD', 'NAME'])
            table.align['NAME'] = 'l'
            table.align['SEED'] = 'r'
            table.align['LEECH'] = 'r'
            table.align['RATIO'] = 'r'
            table.align['SIZE'] = 'r'
            table.align['UPLOAD'] = 'l'

        table.max_width = columns
        table.border = False
        table.padding_width = 1

        for n, result in enumerate(results):
            torrent_name = result['name']

            if local:
                content = [n, result['date'], result['size'],
                           torrent_name[:columns - 42]]
            else:
                no_seeders = int(result['seeders'])
                no_leechers = int(result['leechers'])
                size = result['size']
                date = result['uploaded']

                # compute the S/L ratio (Higher is better)
                try:
                    ratio = no_seeders / no_leechers
                except ZeroDivisionError:
                    ratio = float('inf')

                content = [n, no_seeders, no_leechers,
                           '{:.1f}'.format(ratio),
                           size, date, torrent_name[:columns - 50]]

            if even or not self.enable_color:
                table.add_row(content)
            else:
                table.add_row(content, fore_color='blue')

            # Alternate between colors
            even = not even
        self.print(table)

    def descriptions(self, chosen_links, results, site, timeout):
        for link in chosen_links:
            result = results[link]
            req = request.Request(
                site + '/t.php?id=' + str(result['id']),
                headers=pirate.data.default_headers)
            req.add_header('Accept-encoding', 'gzip')
            f = request.urlopen(req, timeout=timeout)

            if f.info().get('Content-Encoding') == 'gzip':
                f = gzip.GzipFile(fileobj=BytesIO(f.read()))

            res = json.load(f)

            # Replace HTML links with markdown style versions
            desc = re.sub(r'<a href="\s*([^"]+?)\s*"[^>]*>(\s*)([^<]+?)(\s*'
                          r')</a>', r'\2[\3](\1)\4', res['descr'])

            self.print('Description for "{}":'.format(result['name']),
                       color='zebra_1')
            self.print(desc, color='zebra_0')

    def file_lists(self, chosen_links, results, site, timeout):
        # the API may returns object instead of list
        def get(obj):
            try:
                return obj[0]
            except KeyError:
                return obj['0']

        for link in chosen_links:
            result = results[link]
            req = request.Request(
                site + '/f.php?id=' + str(result['id']),
                headers=pirate.data.default_headers)
            req.add_header('Accept-encoding', 'gzip')
            f = request.urlopen(req, timeout=timeout)

            if f.info().get('Content-Encoding') == 'gzip':
                f = gzip.GzipFile(fileobj=BytesIO(f.read()))

            res = json.load(f)

            if len(res) == 1 and 'not found' in get(res[0]['name']):
                self.print('File list not available.')
                return

            self.print('Files in {}:'.format(result['name']), color='zebra_1')
            cur_color = 'zebra_0'

            for f in res:
                name = get(f['name'])
                size = pirate.torrent.pretty_size(int(get(f['size'])))
                self.print('{:>11} {}'.format(
                    size, name),
                    color=cur_color)
                cur_color = 'zebra_0' if cur_color == 'zebra_1' else 'zebra_1'



#!/usr/bin/env python3
import os
import unittest
import json
import sys

from unittest.mock import patch, call, MagicMock
from pirate.print import Printer


class TestPrint(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # needed to display the results table
        os.environ['COLUMNS'] = '80'

    def test_print_results_remote(self):
        class MockTable:
            add_row = MagicMock()
            align = {}
        mock = MockTable()
        printer = Printer(False)
        printer.print = MagicMock()
        with patch('veryprettytable.VeryPrettyTable',
                   return_value=mock) as prettytable:
            results = [{
                'name': 'name',
                'seeders': 1,
                'leechers': 2,
                'size': '3.0 MiB',
                'uploaded': 'never'
            }]
            printer.search_results(results)
            prettytable.assert_called_once_with([
                'LINK', 'SEED', 'LEECH', 'RATIO',
                'SIZE', 'UPLOAD', 'NAME'])
            mock.add_row.assert_has_calls([
                call([0, 1, 2, '0.5', '3.0 MiB', 'never', 'name'])])

    def test_print_results_local(self):
        class MockTable:
            add_row = MagicMock()
            align = {}
        mock = MockTable()
        printer = Printer(False)
        printer.print = MagicMock()
        with patch('veryprettytable.VeryPrettyTable',
                   return_value=mock) as prettytable:
            results = [{
                'name': 'name1',
                'date': '1',
                'size': '1',
            }, {
                'name': 'name2',
                'date': '2',
                'size': '2',
            }]
            printer.search_results(results, local=True)
            prettytable.assert_called_once_with(
                ['LINK', 'DATE', 'SIZE', 'NAME'])
            mock.add_row.assert_has_calls(
                [call([0, '1', '1', 'name1']), call([1, '2', '2', 'name2'])])

    def test_print_color(self):
        printer = Printer(False)
        with patch('pirate.print.builtins.print') as mock_print:
            printer.print('abc', color='zebra_1')
            mock_print.assert_called_once_with(
                'abc',
                file=sys.stderr)
        printer = Printer(True)
        with patch('pirate.print.builtins.print') as mock_print:
            printer.print('abc', color='zebra_1')
            mock_print.assert_called_once_with(
                '\x1b[34mabc', '\x1b[0m',
                file=sys.stderr)

    def test_print_results_local2(self):
        class MockTable:
            add_row = MagicMock()
            align = {}
        mock = MockTable()
        printer = Printer(True)
        printer.print = MagicMock()
        with patch('veryprettytable.VeryPrettyTable',
                   return_value=mock) as prettytable:
            results = [{
                'name': 'name1',
                'date': '1',
                'size': '1',
            }, {
                'name': 'name2',
                'date': '2',
                'size': '2',
            }]
            printer.search_results(results, local=True)
            prettytable.assert_called_once_with(
                ['LINK', 'DATE', 'SIZE', 'NAME'])
            mock.add_row.assert_has_calls([
                call([0, '1', '1', 'name1']),
                call([1, '2', '2', 'name2'], fore_color='blue')])

    def test_print_descriptions(self):
        printer = Printer(False)
        printer.print = MagicMock()

        class MockRequest():
            add_header = MagicMock()
        request_obj = MockRequest()

        class MockResponse():
            read = MagicMock(return_value=json.dumps(
                {'name': 'cool torrent',
                 'descr': 'A fake torrent.\n'}))
            info = MagicMock()
        response_obj = MockResponse()

        with patch('urllib.request.Request', return_value=request_obj):
            with patch('urllib.request.urlopen',
                       return_value=response_obj):
                printer.descriptions([0], [{'id': '1', 'name': 'name'}],
                                     'example.com', 9)
                printer.print.assert_has_calls([
                    call('Description for "name":', color='zebra_1'),
                    call('A fake torrent.\n', color='zebra_0')])

    def test_print_file_lists(self):
        printer = Printer(False)
        printer.print = MagicMock()

        class MockRequest():
            add_header = MagicMock()
            info = MagicMock()
        request_obj = MockRequest()

        class MockResponse():
            read = MagicMock(return_value=json.dumps(
                [{'name': ['readme.txt'], 'size': [16]},
                 {'name': ['a.mkv'], 'size': [677739464]},
                 {'name': ['b.nfo'], 'size': [61]}]))
            info = MagicMock()
        response_obj = MockResponse()

        with patch('urllib.request.Request',
                   return_value=request_obj):
            with patch('urllib.request.urlopen',
                       return_value=response_obj):
                printer.file_lists([0], [{'id': '1', 'name': 'name'}],
                                   'example.com', 9)
                printer.print.assert_has_calls([
                    call('Files in name:', color='zebra_1'),
                    call('       16 B readme.txt', color='zebra_0'),
                    call('  646.3 MiB a.mkv', color='zebra_1'),
                    call('       61 B b.nfo', color='zebra_0')])


if __name__ == '__main__':
    unittest.main()



import re
import sys
import gzip
import pyperclip
import urllib.request as request
import urllib.parse as parse
import urllib.error
import os.path

import pirate.data
import json

from datetime import datetime
from io import BytesIO


def parse_category(printer, category):
    try:
        category = int(category)
    except ValueError:
        pass
    if category in pirate.data.categories.values():
        return category
    elif category in pirate.data.categories.keys():
        return pirate.data.categories[category]
    else:
        printer.print('Invalid category ignored', color='WARN')
        return 0


def parse_sort(printer, sort):
    try:
        sort = int(sort)
    except ValueError:
        pass
    for key, val in pirate.data.sorts.items():
        if sort == key or sort == val[0]:
            return val[1:]
    else:
        printer.print('Invalid sort ignored', color='WARN')
        return pirate.data.sorts['Default'][1:]


def parse_page(page):
    results = []
    try:
        data = json.load(page)
    except json.decoder.JSONDecodeError:
        raise IOError('invalid JSON in API reply: blocked mirror?')

    if len(data) == 1 and 'No results' in data[0]['name']:
        return results

    for res in data:
        res['raw_size'] = int(res['size'])
        res['size'] = pretty_size(int(res['size']))
        res['magnet'] = build_magnet(res['name'], res['info_hash'])
        res['info_hash'] = int(res['info_hash'], 16)
        res['raw_uploaded'] = int(res['added'])
        res['uploaded'] = pretty_date(res['added'])
        res['seeders'] = int(res['seeders'])
        res['leechers'] = int(res['leechers'])
        res['category'] = int(res['category'])
        results.append(res)

    return results


def sort_results(sort, res):
    key, reverse = sort
    return sorted(res, key=lambda x: x[key], reverse=reverse)


def pretty_size(size):
    ranges = [('PiB', 1125899906842624),
              ('TiB', 1099511627776),
              ('GiB', 1073741824),
              ('MiB', 1048576),
              ('KiB', 1024)]
    for unit, value in ranges:
        if size >= value:
            return '{:.1f} {}'.format(size/value, unit)
    return str(size) + ' B'


def pretty_date(ts):
    date = datetime.fromtimestamp(int(ts))
    return date.strftime('%Y-%m-%d %H:%M')


def build_magnet(name, info_hash):
    return 'magnet:?xt=urn:btih:{}&dn={}'.format(
        info_hash, parse.quote(name, ''))


def build_request_path(mode, page, category, terms):
    if mode == 'search':
        query = '/q.php?q={}&cat={}'.format(' '.join(terms), category)
    elif mode == 'top':
        cat = 'all' if category == 0 else category
        query = '/precompiled/data_top100_{}.json'.format(cat)
    elif mode == 'recent':
        query = '/precompiled/data_top100_recent_{}.json'.format(page)
    elif mode == 'browse':
        if category == 0:
            raise Exception('You must specify a category')
        query = '/q.php?q=category:{}'.format(category)
    else:
        raise Exception('Invalid mode', mode)

    return parse.quote(query, '?=&/')


def remote(printer, pages, category, sort, mode, terms, mirror, timeout):
    results = []
    for i in range(1, pages + 1):
        query = build_request_path(mode, i, category, terms)

        # Catch the Ctrl-C exception and exit cleanly
        try:
            req = request.Request(
                mirror + query,
                headers=pirate.data.default_headers)
            try:
                f = request.urlopen(req, timeout=timeout)
            except urllib.error.URLError as e:
                raise e

            if f.info().get('Content-Encoding') == 'gzip':
                f = gzip.GzipFile(fileobj=BytesIO(f.read()))
        except KeyboardInterrupt:
            printer.print('\nCancelled.')
            sys.exit(0)

        results.extend(parse_page(f))

    return sort_results(sort, results)


def find_api(mirror, timeout):
    # try common paths
    for path in ['', '/apip', '/api.php?url=']:
        req = request.Request(mirror + path + '/q.php?q=test&cat=0',
                              headers=pirate.data.default_headers)
        try:
            f = request.urlopen(req, timeout=timeout)
            if f.info().get_content_type() == 'application/json':
                return mirror + path
        except urllib.error.URLError as e:
            res = e.fp.read().decode()
            if e.code == 503 and 'cf-browser-verification' in res:
                raise IOError('Cloudflare protected')

    # extract api path from main.js
    req = request.Request(mirror + '/static/main.js',
                          headers=pirate.data.default_headers)
    try:
        f = request.urlopen(req, timeout=timeout)
        if f.info().get_content_type() == 'application/javascript':
            match = re.search("var server='([^']+)'", f.read().decode())
            return mirror + match.group(1)
    except urllib.error.URLError:
        raise IOError('API not found: no main.js')

    raise IOError('API not found')


def get_torrent(info_hash, timeout):
    url = 'http://itorrents.org/torrent/{:X}.torrent'
    req = request.Request(url.format(info_hash),
                          headers=pirate.data.default_headers)
    req.add_header('Accept-encoding', 'gzip')

    torrent = request.urlopen(req, timeout=timeout)
    if torrent.info().get('Content-Encoding') == 'gzip':
        torrent = gzip.GzipFile(fileobj=BytesIO(torrent.read()))

    return torrent.read()


def save_torrents(printer, chosen_links, results, folder, timeout):
    for link in chosen_links:
        result = results[link]
        torrent_name = result['name'].replace('/', '_').replace('\\', '_')
        file = os.path.join(folder, torrent_name + '.torrent')

        try:
            torrent = get_torrent(result['info_hash'], timeout)
        except urllib.error.HTTPError as e:
            printer.print('There is no cached file for this torrent :('
                          ' \nCode: {} - {}'.format(e.code, e.reason),
                          color='ERROR')
        else:
            open(file, 'wb').write(torrent)
            printer.print('Saved {:X} in {}'.format(result['info_hash'], file))


def save_magnets(printer, chosen_links, results, folder):
    for link in chosen_links:
        result = results[link]
        torrent_name = result['name'].replace('/', '_').replace('\\', '_')
        file = os.path.join(folder,  torrent_name + '.magnet')

        printer.print('Saved {:X} in {}'.format(result['info_hash'], file))
        with open(file, 'w') as f:
            f.write(result['magnet'] + '\n')


def copy_magnets(printer, chosen_links, results):
    clipboard_text = ''
    for link in chosen_links:
        result = results[link]
        clipboard_text += result['magnet'] + "\n"
        printer.print('Copying {:X} to clipboard'.format(result['info_hash']))

    pyperclip.copy(clipboard_text)
